/** emacs: this is -*- c++ -*- **/
/**
 **   @file    histogram.h        
 **                   
 **   @author  sutt
 **   @date    Mon 28 Dec 2020 19:33:12 GMT
 **
 **   $Id: histogram.h, v0.0   Mon 28 Dec 2020 19:33:12 GMT sutt $
 **
 **   Copyright (C) 2020 sutt (sutt@cern.ch)    
 **
 **/


#ifndef  HISTOGRAM_H
#define  HISTOGRAM_H

#include <string>
#include <vector>
#include <cmath>
#include <iostream>
#include <exception>


#include "serialise_base.h"
#include "binary_search.h"



class histogram {

public:

  // histogram error exception
  class exception : public std::exception { 
  public:
    exception(const std::string& s) { std::cerr << what() << " " << s << std::endl; }; 
    virtual const char* what() const throw() { return ""; }
  };

public:

  histogram( const std::string& name="" ) : mname(name) { } 

  histogram( const std::string& name, size_t nbins, const double* limits ) : mname(name) {
    if ( nbins>0 ) create( nbins, limits );
    else throw exception("histogram: not enough bins creating histogram: "+name);
  }

  histogram( const std::string& name, const std::vector<double>& limits ) : mname(name) {
    if ( limits.size()>1 ) create( limits.size()-1, &limits[0] );
    else throw exception("histogram: not enough bin limits creating histogram: "+name);
  }

  histogram( const std::string& name, size_t nbins, double lo, double hi ) : mname(name) {    
    if ( nbins==0 ) throw exception("histogram: not enough bins creating histogram: "+name);
    std::vector<double> limits(nbins+1,0);
    for ( size_t i=nbins ; i-- ; ) limits[i] = (lo*(nbins-i) + hi*i)/nbins;
    /// make sure the limits at least are exact 
    limits[0]     = lo;
    limits[nbins] = hi;
    create( nbins, &limits[0] );
  } 

  /// constructor from stream

  histogram( const std::vector<double> stream ) { 
    deserialise( stream );
  }

  virtual ~histogram() { } 

  histogram( const histogram& h ) : 
    mname(h.mname),
    mxlimits(h.mxlimits),
    mx(h.mx),
    my(h.my),
    mye(h.mye) 
  { }


  histogram& operator=( histogram&& h ) {
    std::swap( mname, h.mname );
  
    std::swap( mxlimits, h.mxlimits );
    std::swap( mx, h.mx );

    std::swap( my,  h.my );
    std::swap( mye, h.mye );
    
    return *this;
  }


  histogram& operator=( const histogram& h ) {
    mname    = h.mname;
    mxlimits = h.mxlimits;
    mx       = h.mx;
    my       = h.my;
    mye      = h.mye;
    return *this;
  }

  /// accessors 

  std::string name() const { return mname; }

  size_t size() const { return mx.size(); }

  double  operator[]( int i ) const { return my.at(i); }
  double& operator[]( int i )       { return my[i]; }

  double&  y( int i ) { return my.at(i);  }
  double& ye( int i ) { return mye.at(i); }

  double  y( int i ) const { return my.at(i);  }
  double ye( int i ) const { return mye.at(i); }

  double lo( int i ) const { return mxlimits.at(i); }
  double hi( int i ) const { return mxlimits.at(i+1); }

  const std::vector<double>& xlimits() const { return mxlimits; } 

  /// logical operators - 
  /// NB: these only test on the BIN LIMITS, NOT THE BIN CONTENTS !!
  
  bool operator==( const histogram& h ) {
    if ( size()!=h.size() ) return false;
    if ( size()==0 ) return true;
    double delta = mxlimits[1]-mxlimits[0];
    for ( size_t i=mxlimits.size() ; i-- ; ) if ( std::fabs(mxlimits[i]-h.mxlimits[i])>delta*1e-10 ) return false;
    return true;
  }

  bool operator!=( const histogram& h ) { 
    return !(operator==(h));
  }

  /// mathematical operations ...

  histogram& operator*=( double d ) { 
    for ( size_t i=size() ; i-- ; ) { 
      my[i]  *= d;
      mye[i] *= d;
    }
    return *this;
  }

  histogram operator*( double d ) const {
    histogram h(*this);
    return (h*=d);
  }

  histogram& operator/=( double d ) {
    return operator*=(1/d);
  }

  histogram operator/( double d ) const {
    histogram h(*this);
    return h/=d;
  }
  

  histogram& operator+=( const histogram& h ) { 
    if ( ! operator==(h) )  throw exception( "histogram: bin mismatch for operator +- " );
    for ( size_t i=size() ; i-- ; ) { 
      my[i]  += h.my[i];
      mye[i]  = std::sqrt( h.mye[i]*h.mye[i] + mye[i]*mye[i] );
    }
    return *this;
  }

  histogram operator+( histogram h ) const { 
    return ( h += (*this) );
  }


  histogram operator-=( const histogram& h ) { 
    return operator+=( h*(-1) );
  }

  histogram operator-( const histogram& h ) const { 
    histogram hi(*this);
    return hi-=h;
  }
  

  /// filling ...

  int index( double x ) const { return find_index( x, mxlimits ); }

  
  void fill( double x, double w=1 ) {
    int ibin = index( x );
    if ( ibin != -1 ) { 
      my[ibin] += w;
      mye[ibin] = std::sqrt( mye[ibin]*mye[ibin] + w );
    } 
  }
  

  /// serialisation and deserialisation ...

  std::vector<double> serialise() const { 
   
    std::vector<double> s;
    
    s.push_back( SB::WRITEGUARD );

    s.push_back(0); // overall size

    SB::serialise( s, name() );
    SB::serialise( s, mxlimits );
    /// don't need to save this - can compute from the limits
    /// SB::serialise( s, mx );
    SB::serialise( s, my );
    SB::serialise( s, mye );

    s.push_back( SB::WRITEGUARD );

    s[1] = s.size();

    return s;
  }


  void deserialise( const std::vector<double>& v ) { 
 
    std::vector<double>::const_iterator itr = v.begin();
    
    if ( (*itr++) != SB::WRITEGUARD ) throw exception("histogram: read error");
   
    /// don't need this now ...
    //  int ntotal = (*itr++); 
    (*itr++);

    SB::deserialise( itr, mname );
    SB::deserialise( itr, mxlimits );
    /// SB::deserialise( itr, mx );
    SB::deserialise( itr, my );
    SB::deserialise( itr, mye );
    
    double var = (*itr++);

    //    std::cout << "read: " << mname << "\tsize: " << ntotal << std::endl;
    
    if ( var != SB::WRITEGUARD ) throw exception("histogram: read error reading histogram: "+name());

    mx.resize(my.size());
    for ( size_t i=my.size() ; i-- ; ) mx[i] = 0.5*(mxlimits[i]+mxlimits[i+1]);
   
  } 

protected:

  void create( size_t nbins, const double* limits ) { 
    
    mxlimits.resize(nbins+1);
    
    for ( size_t i=nbins+1 ; i-- ; ) mxlimits[i] = limits[i];
    
    mx.resize(nbins);
    
    for ( size_t i=nbins ; i-- ; ) mx[i] = 0.5*(limits[i]+limits[i+1]);
    
    my  = std::vector<double>(nbins,0);
    mye = std::vector<double>(nbins,0);
    
  }
 
protected:
  
  std::string mname;
  
  std::vector<double> mxlimits;
  
  std::vector<double> mx;
  
  std::vector<double> my;
  std::vector<double> mye;
  
};


inline histogram operator*( double d, histogram h ) {
  return (h*=d);
}

  


inline std::ostream& operator<<( std::ostream& s, const histogram& h ) { 
  s << "key: " << h.name() << std::endl;
  for ( size_t i=0 ; i<h.size() ; i++ ) s << "  " << h.lo(i) << "\t- " << h.hi(i) << "\t: " << h.y(i) << "\t+- " << h.ye(i) << std::endl; 
  return s;
}


#endif  // HISTOGRAM_H 










